from .device_manager import DeviceManagerProcess
from .system_identity import SystemIdentityProcess
from .system_stats import SystemStatsProcess

__all__ = [
    "DeviceManagerProcess",
    "SystemIdentityProcess",
    "SystemStatsProcess",
]



from bussdcc import RuntimeProtocol

from . import graph


class DeviceReconciler:
    def reconcile(
        self,
        runtime: RuntimeProtocol,
        desired_nodes: dict[str, graph.DeviceNode],
    ) -> None:
        actual = {d.id: d for d in runtime.devices.list()}
        desired_ids = set(desired_nodes.keys())
        actual_ids = set(actual.keys())

        order = graph.topo_sort(desired_nodes)
        reverse_order = list(reversed(order))
        dependents = graph.build_dependents(desired_nodes)

        deleted_ids = actual_ids - desired_ids

        dirty_ids = graph.initial_dirty_ids(runtime, desired_nodes)
        dirty_ids = graph.expand_dirty_ids(dirty_ids, dependents)

        # detach deleted devices first in reverse dependency order
        for dev_id in graph.deleted_detach_order(runtime, deleted_ids):
            try:
                runtime.devices.detach(dev_id)
            except Exception:
                pass

        # detach changed/new dependency subgraphs in reverse topo order
        for dev_id in reverse_order:
            if dev_id not in dirty_ids:
                continue

            if runtime.devices.get(dev_id) is None:
                continue

            try:
                runtime.devices.detach(dev_id)
            except Exception:
                pass

        # attach changed/new nodes in topo order
        for dev_id in order:
            if dev_id not in dirty_ids:
                continue

            try:
                runtime.devices.attach(desired_nodes[dev_id].device)
            except Exception:
                pass



from typing import Any

from flask import Blueprint, render_template, redirect, url_for, request

from bussdcc import ContextProtocol
from bussdcc_framework.web import FlaskApp, WebPlugin


class SystemServicesPlugin:
    name = "system-services"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_system_services",
            __name__,
            url_prefix="/system/services",
            template_folder="templates",
        )

        @bp.route("/")
        def index() -> Any:
            supervisor = ctx.runtime.services
            services = supervisor.statuses()

            return render_template(
                "bussdcc_system/services/index.html", services=services
            )

        @bp.route("/stop/<service>")
        def stop(service: str) -> Any:
            supervisor = ctx.runtime.services
            supervisor.stop(service)

            return redirect(url_for("bussdcc_system_services.index"))

        @bp.route("/start/<service>")
        def start(service: str) -> Any:
            supervisor = ctx.runtime.services
            supervisor.start(service)

            return redirect(url_for("bussdcc_system_services.index"))

        app.register_blueprint(bp)


plugin: WebPlugin = SystemServicesPlugin()



from .device import (
    DeviceSpec,
    DeviceMap,
    replace_devices,
    add_device,
    update_device_config,
    delete_device,
)

__all__ = [
    "DeviceSpec",
    "DeviceMap",
    "replace_devices",
    "add_device",
    "update_device_config",
    "delete_device",
]



from typing import Any
from flask import Blueprint, render_template, flash, redirect, url_for, request

from bussdcc import ContextProtocol
from bussdcc_framework.web import FlaskApp, WebPlugin
from bussdcc_framework.interface.web import current_ctx
from bussdcc_framework.codec import load_value, dump_value
from bussdcc_framework.interface.web import formtree
from bussdcc_hardware.registry import registry

from bussdcc_system.service.device_manager.graph import extract_dependencies
from bussdcc_system.model import DeviceSpec

from .... import message


class SystemDevicesPlugin:
    name = "system-devices"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_system_devices",
            __name__,
            url_prefix="/system/devices",
            template_folder="templates",
        )

        @bp.route("/")
        def index() -> Any:
            ctx = current_ctx()
            devices = ctx.state.get("devices", {})

            runtime_devices = {dev.id: dev for dev in ctx.runtime.devices.list()}
            online_status = {dev.id: dev.online for dev in runtime_devices.values()}

            dependency_map: dict[str, set[str]] = {}
            dependent_map: dict[str, set[str]] = {
                device_id: set() for device_id in devices
            }
            missing_dependency_map: dict[str, set[str]] = {}
            status_map: dict[str, str] = {}

            configured_ids = set(devices.keys())

            for device_id, spec in devices.items():
                registry_entry = registry.devices.get(spec.type)

                if registry_entry is None or registry_entry.definition is None:
                    dependency_map[device_id] = set()
                    missing_dependency_map[device_id] = set()
                    status_map[device_id] = "type-unavailable"
                    continue

                definition = registry_entry.definition
                device_cfg = load_value(definition.config_class, spec.config)
                deps = extract_dependencies(device_cfg)

                dependency_map[device_id] = deps
                missing = {dep for dep in deps if dep not in configured_ids}
                missing_dependency_map[device_id] = missing

                for dep in deps:
                    if dep in dependent_map:
                        dependent_map[dep].add(device_id)

                if missing:
                    status_map[device_id] = "missing-dependency"
                elif online_status.get(device_id):
                    status_map[device_id] = "online"
                else:
                    status_map[device_id] = "offline"

            roots = sorted(
                device_id
                for device_id, deps in dependency_map.items()
                if not deps or any(dep not in configured_ids for dep in deps)
            )

            return render_template(
                "bussdcc_system/devices/index.html",
                devices=devices,
                dependency_map=dependency_map,
                dependent_map={k: sorted(v) for k, v in dependent_map.items()},
                missing_dependency_map=missing_dependency_map,
                status_map=status_map,
                roots=roots,
            )

        @bp.route("/new")
        def new() -> Any:
            devices = registry.devices
            name = request.args.get("name")
            type_ = request.args.get("type_")
            if name and type_:
                registry_entry = registry.devices[type_]
                definition = registry_entry.definition
                if definition is None:
                    flash("Device not available", "warning")
                    return redirect(url_for("bussdcc_system_devices.index"))

                tree = formtree.build(definition.config_class)
                return render_template(
                    "bussdcc_system/devices/new_config.html",
                    name=name,
                    type=type_,
                    tree=tree,
                    action=url_for(
                        "bussdcc_system_devices.create", name=name, type_=type_
                    ),
                )
            else:
                return render_template(
                    "bussdcc_system/devices/new.html", devices=devices
                )

        @bp.route("/show/<id>")
        def show(id: str) -> Any:
            ctx = current_ctx()
            devices = ctx.state.get("devices", {})
            spec = devices.get(id)

            if spec is None:
                flash("Device not found", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            registry_entry = registry.devices.get(spec.type)
            if registry_entry is None or registry_entry.definition is None:
                flash("Device type not available", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            definition = registry_entry.definition
            device_cfg = load_value(definition.config_class, spec.config)
            tree = formtree.build(device_cfg)

            return render_template(
                "bussdcc_system/devices/show.html",
                tree=tree,
                device_id=id,
                action=url_for("bussdcc_system_devices.update", id=id),
            )

        @bp.route("/create/<type_>/<name>", methods=["POST"])
        def create(type_: str, name: str) -> Any:
            ctx = current_ctx()

            registry_entry = registry.devices[type_]
            definition = registry_entry.definition
            if definition is None:
                flash("Device not available", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            tree = formtree.build(definition.config_class)
            data = formtree.unflatten(tree, request.form)
            cfg = load_value(definition.config_class, data)
            ctx.emit(
                message.DeviceAdded(
                    device=name,
                    spec=DeviceSpec(
                        type=type_,
                        config=dump_value(cfg),
                    ),
                )
            )

            return redirect(url_for("bussdcc_system_devices.index"))

        @bp.route("/update/<id>", methods=["POST"])
        def update(id: str) -> Any:
            ctx = current_ctx()
            devices = ctx.state.get("devices", {})
            spec = devices.get(id)

            if spec is None:
                flash("Device not found", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            registry_entry = registry.devices.get(spec.type)
            if registry_entry is None or registry_entry.definition is None:
                flash("Device type not available", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            definition = registry_entry.definition
            tree = formtree.build(definition.config_class)
            tree = formtree.validate(tree, request.form)

            if tree.errors > 0:
                return render_template(
                    "bussdcc_system/devices/show.html",
                    id=id,
                    spec=spec,
                    definition=definition,
                    tree=tree,
                )

            data = formtree.unflatten(tree, request.form)
            device_cfg = load_value(definition.config_class, data)

            ctx.emit(
                message.DeviceConfigUpdate(
                    device=id,
                    config=dump_value(device_cfg),
                )
            )

            flash("Device updated", "success")
            return redirect(url_for("bussdcc_system_devices.index"))

        @bp.route("/delete/<id>", methods=["POST"])
        def delete(id: str) -> Any:
            ctx = current_ctx()
            devices = ctx.state.get("devices", {})
            spec = devices.get(id)

            if spec is None:
                flash("Device not found", "warning")
                return redirect(url_for("bussdcc_system_devices.index"))

            ctx.emit(message.DeviceDeleted(device=id))
            flash(f"Deleted device '{id}'", "success")

            return redirect(url_for("bussdcc_system_devices.index"))

        app.register_blueprint(bp)


plugin: WebPlugin = SystemDevicesPlugin()



from bussdcc import ContextProtocol, Event, Message, Service
from bussdcc_hardware.registry import registry

from ... import message
from .graph import build_desired_nodes
from .reconciler import DeviceReconciler


class DeviceManagerService(Service):
    name = "device_manager"

    def __init__(self) -> None:
        self.reconciler = DeviceReconciler()

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        if isinstance(
            evt.payload,
            (
                message.DevicesReplaced,
                message.DeviceAdded,
                message.DeviceConfigUpdate,
                message.DeviceDeleted,
            ),
        ):
            devices = ctx.state.get("devices", {})
            desired_nodes = build_desired_nodes(registry.devices, devices)
            self.reconciler.reconcile(ctx.runtime, desired_nodes)



from typing import Any, Mapping
from dataclasses import dataclass, fields, is_dataclass

from bussdcc import RuntimeProtocol, DeviceProtocol
from bussdcc_framework.codec import load_value
from bussdcc_framework.metadata import FieldMetadata

from ...model import DeviceSpec


@dataclass(slots=True)
class GraphNode:
    id: str
    device: DeviceProtocol[Any]
    deps: set[str]


@dataclass(slots=True)
class DeviceNode(GraphNode):
    type_: str
    spec: DeviceSpec


def extract_dependencies(cfg: Any) -> set[str]:
    deps: set[str] = set()

    if not is_dataclass(cfg):
        return deps

    for f in fields(cfg):
        meta = FieldMetadata.from_field(f)
        if meta.ref is None:
            continue

        value = getattr(cfg, f.name)

        if value is None:
            continue

        if isinstance(value, str):
            if value:
                deps.add(value)
            continue

        if isinstance(value, list):
            for item in value:
                if isinstance(item, str) and item:
                    deps.add(item)
            continue

        if isinstance(value, dict):
            for item in value.values():
                if isinstance(item, str) and item:
                    deps.add(item)
            continue

    return deps


def build_dependents(nodes: Mapping[str, GraphNode]) -> dict[str, set[str]]:
    dependents: dict[str, set[str]] = {node_id: set() for node_id in nodes}

    for node_id, node in nodes.items():
        for dep in node.deps:
            if dep in nodes:
                dependents[dep].add(node_id)

    return dependents


def topo_sort(nodes: Mapping[str, GraphNode]) -> list[str]:
    in_degree: dict[str, int] = {node_id: 0 for node_id in nodes}
    dependents: dict[str, set[str]] = {node_id: set() for node_id in nodes}

    for node_id, node in nodes.items():
        for dep in node.deps:
            if dep not in nodes:
                continue
            in_degree[node_id] += 1
            dependents[dep].add(node_id)

    ready = sorted(node_id for node_id, deg in in_degree.items() if deg == 0)
    result: list[str] = []

    while ready:
        node_id = ready.pop(0)
        result.append(node_id)

        for child in sorted(dependents[node_id]):
            in_degree[child] -= 1
            if in_degree[child] == 0:
                ready.append(child)

    if len(result) != len(nodes):
        raise RuntimeError("Device dependency cycle detected")

    return result


def initial_dirty_ids(
    runtime: RuntimeProtocol,
    nodes: Mapping[str, GraphNode],
) -> set[str]:
    actual = {d.id: d for d in runtime.devices.list()}
    dirty: set[str] = set()

    for node_id, node in nodes.items():
        actual_dev = actual.get(node_id)
        if actual_dev is None:
            dirty.add(node_id)
            continue

        if type(actual_dev) is not type(node.device):
            dirty.add(node_id)
            continue

        if actual_dev.config != node.device.config:
            dirty.add(node_id)
            continue

        if any(dep not in nodes for dep in node.deps):
            dirty.add(node_id)
            continue

        if any(dep not in actual for dep in node.deps):
            dirty.add(node_id)
            continue

    return dirty


def expand_dirty_ids(
    dirty: set[str],
    dependents: Mapping[str, set[str]],
) -> set[str]:
    expanded = set(dirty)
    stack = list(dirty)

    while stack:
        node_id = stack.pop()

        for child in dependents.get(node_id, set()):
            if child in expanded:
                continue
            expanded.add(child)
            stack.append(child)

    return expanded


def build_desired_nodes(
    col: dict[str, Any],
    devices_cfg: dict[str, DeviceSpec],
) -> dict[str, DeviceNode]:
    desired: dict[str, DeviceNode] = {}

    for device_id, spec in devices_cfg.items():
        entry = col.get(spec.type)

        if not entry or not entry.available:
            continue

        definition = entry.definition
        if definition is None:
            continue

        cfg = load_value(definition.config_class, spec.config)
        device = definition.driver_class(id=device_id, config=cfg)
        deps = extract_dependencies(cfg)

        desired[device_id] = DeviceNode(
            id=device_id,
            type_=spec.type,
            spec=spec,
            device=device,
            deps=deps,
        )

    return desired


def build_actual_nodes(runtime: RuntimeProtocol) -> dict[str, GraphNode]:
    actual: dict[str, GraphNode] = {}

    for device in runtime.devices.list():
        actual[device.id] = GraphNode(
            id=device.id,
            device=device,
            deps=extract_dependencies(device.config),
        )

    return actual


def deleted_detach_order(
    runtime: RuntimeProtocol,
    deleted_ids: set[str],
) -> list[str]:
    actual_nodes = build_actual_nodes(runtime)
    deleted_nodes = {
        node_id: node
        for node_id, node in actual_nodes.items()
        if node_id in deleted_ids
    }
    return list(reversed(topo_sort(deleted_nodes)))



from dataclasses import dataclass
from typing import Any


@dataclass(slots=True, frozen=True)
class DeviceSpec:
    type: str
    config: dict[str, Any]


DeviceMap = dict[str, DeviceSpec]


def replace_devices(devices: DeviceMap) -> DeviceMap:
    return dict(devices)


def add_device(
    devices: DeviceMap,
    device: str,
    spec: DeviceSpec,
) -> DeviceMap:
    return {
        **devices,
        device: spec,
    }


def update_device_config(
    devices: DeviceMap,
    device: str,
    config: dict[str, object],
) -> DeviceMap:
    spec = devices[device]

    return {
        **devices,
        device: DeviceSpec(
            type=spec.type,
            config=config,
        ),
    }


def delete_device(
    devices: DeviceMap,
    device: str,
) -> DeviceMap:
    return {
        device_id: spec for device_id, spec in devices.items() if device_id != device
    }



from typing import Any

from flask import Blueprint, render_template, redirect, url_for, request

from bussdcc import ContextProtocol
from bussdcc_framework.web import FlaskApp, WebPlugin


class SystemRuntimeInfoPlugin:
    name = "system-runtime-info"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_system_runtime_info",
            __name__,
            url_prefix="/system/runtime/info",
            template_folder="templates",
        )

        @bp.route("/")
        def index() -> Any:
            runtime_info = ctx.state.get("runtime_info", {})

            return render_template(
                "bussdcc_system/runtime_info/index.html", runtime_info=runtime_info
            )

        app.register_blueprint(bp)


plugin: WebPlugin = SystemRuntimeInfoPlugin()



from flask import Blueprint, render_template

from bussdcc import ContextProtocol

from bussdcc_framework.web import FlaskApp, WebPlugin


class SystemStatsPlugin:
    name = "system-stats"

    def init_app(self, app: FlaskApp, ctx: ContextProtocol) -> None:
        bp = Blueprint(
            "bussdcc_system_stats",
            __name__,
            url_prefix="/system/stats",
            template_folder="templates",
        )

        @bp.route("/")
        def index() -> str:
            cpu_history = ctx.state.get("system.cpu.history", {})
            cpu_temperature = ctx.state.get("system.cpu.temperature")
            cpu_usage = ctx.state.get("system.cpu.usage")
            disk_usage = ctx.state.get("system.disk.usage", {"used": 0, "total": 0})
            memory_usage = ctx.state.get("system.memory.usage", {"used": 0, "total": 0})
            network_usage = ctx.state.get("system.network.usage")
            network_history = ctx.state.get("system.network.history", {})
            system_load = ctx.state.get("system.load")

            return render_template(
                "bussdcc_system/stats/index.html",
                cpu_history=cpu_history,
                cpu_temperature=cpu_temperature,
                cpu_usage=cpu_usage,
                disk_usage=disk_usage,
                memory_usage=memory_usage,
                network_usage=network_usage,
                network_history=network_history,
                system_load=system_load,
            )

        app.register_blueprint(bp)


plugin: WebPlugin = SystemStatsPlugin()



from bussdcc import ContextProtocol, Event, Message, Process

from .. import message
from ..model.device import (
    add_device,
    update_device_config,
    delete_device,
    replace_devices,
)


class DeviceManagerProcess(Process):
    name = "device_manager"

    def handle_event(self, ctx: ContextProtocol, evt: Event[Message]) -> None:
        payload = evt.payload

        if isinstance(payload, message.DevicesReplaced):
            ctx.state.set("devices", replace_devices(payload.devices))
            return

        if isinstance(payload, message.DeviceAdded):
            ctx.state.update(
                "devices",
                lambda devices: add_device(
                    devices or {},
                    payload.device,
                    payload.spec,
                ),
            )
            return

        if isinstance(payload, message.DeviceConfigUpdate):
            ctx.state.update(
                "devices",
                lambda devices: update_device_config(
                    devices or {},
                    payload.device,
                    payload.config,
                ),
            )
            return

        if isinstance(payload, message.DeviceDeleted):
            ctx.state.update(
                "devices",
                lambda devices: delete_device(
                    devices or {},
                    payload.device,
                ),
            )



from dataclasses import dataclass

from bussdcc import Message

from ..model import DeviceSpec


@dataclass(slots=True, frozen=True)
class DevicesReplaced(Message):
    devices: dict[str, DeviceSpec]


@dataclass(slots=True, frozen=True)
class DeviceAdded(Message):
    device: str
    spec: DeviceSpec


@dataclass(slots=True, frozen=True)
class DeviceConfigUpdate(Message):
    device: str
    config: dict[str, object]


@dataclass(slots=True, frozen=True)
class DeviceDeleted(Message):
    device: str
